/*
 * Copyright 2012 Jared Bunting, Peach Jean Solutions <jared.bunting@peachjean.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
package net.peachjean.restmx.console.providers;

import net.peachjean.restmx.console.model.MxModel;
import net.peachjean.restmx.console.resource.EncoderBean;
import net.peachjean.restmx.type.HtmlContent;
import net.sf.json.JSONSerializer;

import javax.management.MBeanAttributeInfo;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.ext.MessageBodyWriter;
import javax.ws.rs.ext.Provider;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.util.*;

import static net.peachjean.restmx.console.providers.ProviderUtils.determineCharset;

@Provider
@Produces(MediaType.APPLICATION_JSON)
public class JsonItemProvider implements MessageBodyWriter<MxModel> {
	@Override
	public boolean isWriteable(final Class<?> type, final Type genericType, final Annotation[] annotations,
	                           final MediaType mediaType) {
		return MxModel.class.isAssignableFrom(type) && MediaType.APPLICATION_JSON_TYPE.isCompatible(mediaType);
	}

	@Override
	public long getSize(final MxModel model, final Class<?> type, final Type genericType,
	                    final Annotation[] annotations,
	                    final MediaType mediaType) {
		return -1;
	}

	@Override
	public void writeTo(final MxModel model, final Class<?> type, final Type genericType,
	                    final Annotation[] annotations,
	                    final MediaType mediaType, final MultivaluedMap<String, Object> httpHeaders,
	                    final OutputStream entityStream)
			throws IOException, WebApplicationException {
		// Translate known models, needs a refactoring to embed that in each resource...
		HashMap<String, Object> result = new HashMap<String, Object>();

		result.put("label", model.getLabel());

		boolean leaf = model.isLeaf();

		if (model.getModel().containsKey("items") && !leaf) {
			Object items = model.getModel().get("items");

			Collection<Object> itemCollection = null;
			if (items instanceof Collection) {
				itemCollection = (Collection<Object>) items;
			} else {
				itemCollection = Arrays.asList(items);
			}
			List<Map<String, String>> children = new ArrayList<Map<String, String>>();
			for (Object item : itemCollection) {

				HashMap<String, String> ref = new HashMap<String, String>();

				if (item instanceof MBeanAttributeInfo) {
					ref.put("$ref", new EncoderBean().encode(((MBeanAttributeInfo) item).getName()) + "/");
				} else if (item instanceof Map && ((Map) item).containsKey("declaration")) {
					ref.put("$ref", ((Map) item).get("declaration").toString());
				} else {
					ref.put("$ref", item.toString() + "/");
				}
				children.add(ref);
			}
			result.put("children", children);
		} else {
			if (model.getModel().containsKey("value")) {
				if (model.getModel().get("value") instanceof HtmlContent) {
					result.put("value", "...");
				} else {
					result.put("value", model.getModel().get("value").toString());
				}
			} else if (model.getModel().containsKey("items")) {
				Object items = model.getModel().get("items");
				String value = null;
				if (items.getClass().isArray()) {
					value = Arrays.deepToString((Object[]) items);
				} else {
					value = items.toString();
				}
				result.put("value", value);
			}
		}

		// Hack because root must be a list for dojo tree...
		if ("servers".equals(model.getLabel())) {
			writeResult(mediaType, entityStream, new Object[]{result});
		} else {
			writeResult(mediaType, entityStream, result);
		}
	}

	private void writeResult(final MediaType mediaType, final OutputStream entityStream, final Object result1)
			throws IOException {
		JSONSerializer.toJSON(result1).write(new OutputStreamWriter(entityStream, determineCharset(mediaType))).close();
	}
}
