/*
 * Copyright 2012 Jared Bunting, Peach Jean Solutions <jared.bunting@peachjean.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
package net.peachjean.restmx.console.providers;

import net.peachjean.restmx.console.model.MxModel;
import net.peachjean.restmx.console.resource.EncoderBean;
import org.apache.velocity.Template;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader;

import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.*;
import javax.ws.rs.ext.MessageBodyWriter;
import javax.ws.rs.ext.Provider;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import static net.peachjean.restmx.console.providers.ProviderUtils.determineCharset;

@Provider
@Produces({MediaType.TEXT_PLAIN, MediaType.TEXT_HTML})
public class TemplateItemProvider implements MessageBodyWriter<MxModel> {
	@Context
	private Request request;

	@Context
	private UriInfo uriInfo;

	@Override
	public boolean isWriteable(final Class<?> type, final Type genericType, final Annotation[] annotations,
	                           final MediaType mediaType) {
		return MxModel.class.isAssignableFrom(type) && (MediaType.TEXT_PLAIN_TYPE.isCompatible(mediaType) ||
				MediaType.TEXT_HTML_TYPE.isCompatible(mediaType));
	}

	@Override
	public long getSize(final MxModel mxModel, final Class<?> type, final Type genericType,
	                    final Annotation[] annotations, final MediaType mediaType) {
		return -1;
	}

	@Override
	public void writeTo(final MxModel model, final Class<?> type, final Type genericType,
	                    final Annotation[] annotations, final MediaType mediaType,
	                    final MultivaluedMap<String, Object> httpHeaders, final OutputStream entityStream)
			throws IOException, WebApplicationException {
		httpHeaders.add("Cache-Control", "no-store, no-cache, must-revalidate");
		httpHeaders.add("Cache-Control", "post-check=0, pre-check=0");
		httpHeaders.add("Pragma", "no-cache");
		httpHeaders.add("Expires", "0");

		Map<String, Object> enrichedModel = new HashMap<String, Object>(model.getModel());

		Template template;
		try {
			VelocityEngine ve = new VelocityEngine();

			Properties p = new Properties();
			p.setProperty("resource.loader", "class");
			p.setProperty("class.resource.loader.description",
					"Velocity Classpath Resource Loader");
			p.setProperty("class.resource.loader.class", ClasspathResourceLoader.class.getName());

			ve.init(p);

			template = ve.getTemplate(getTemplateName(model.getTemplateName(), mediaType));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		String skin = null;
		if (skin == null) {
			skin = "default";
		}
		String desc = null;
		if (desc == null) {
			desc = "on";
		}
		enrichedModel.put("query", uriInfo.getPath());
		enrichedModel.put("margin", "embedded".equals(skin) ? 0 : 5);
		enrichedModel.put("skin", skin);
		enrichedModel.put("desc", desc);

		enrichedModel.put("encoder", new EncoderBean());
		enrichedModel.put("request", request);
		enrichedModel.put("label", model.getLabel());

		final OutputStreamWriter writer = new OutputStreamWriter(entityStream, determineCharset(mediaType));
		template.merge(new MapContext(enrichedModel), writer);
		writer.flush();
	}

	private String getTemplateName(final String templateName, final MediaType mediaType) {
		if (MediaType.TEXT_PLAIN_TYPE.isCompatible(mediaType)) {
			return "restmx/templates/" + templateName + "-plain.vm";
		} else if (MediaType.TEXT_HTML_TYPE.isCompatible(mediaType)) {
			return "restmx/templates/" + templateName + ".vm";
		} else {
			throw new IllegalArgumentException("MediaType " + mediaType + " is not supported.");
		}
	}

	private class MapContext implements org.apache.velocity.context.Context {
		private final Map<String, Object> enrichedModel;

		public MapContext(
				final Map<String, Object> enrichedModel) {
			this.enrichedModel = enrichedModel;
		}

		@Override
		public Object put(final String s, final Object o) {
			return this.enrichedModel.put(s, o);
		}

		@Override
		public Object get(final String s) {
			return this.enrichedModel.get(s);
		}

		@Override
		public boolean containsKey(final Object o) {
			return this.enrichedModel.containsKey(o);
		}

		@Override
		public Object[] getKeys() {
			return this.enrichedModel.keySet().toArray();
		}

		@Override
		public Object remove(final Object o) {
			return this.enrichedModel.remove(o);
		}
	}
}
