/* 
 * Copyright 2012 Jared Bunting, Peach Jean Solutions <jared.bunting@peachjean.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */

package net.peachjean.restmx.console.resource;

import net.peachjean.restmx.console.model.MxItemModel;
import net.peachjean.restmx.console.model.MxListModel;
import net.peachjean.restmx.console.model.MxModel;
import net.peachjean.restmx.server.ServerConnectionProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.management.*;
import javax.management.openmbean.CompositeData;
import javax.management.openmbean.TabularData;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.io.IOException;
import java.util.*;

@Path("/servers/{server}/domains/{domain}/mbeans/{mbean}/attributes")
public class AttributeResource {

	private static Logger log = LoggerFactory.getLogger(AttributeResource.class);

	private String templateName = "attribute";

	@Context
	private ServerConnectionProvider serverConnectionProvider;

	@Context
	private UriInfo uriInfo;

	@GET
	@Produces({MediaType.TEXT_HTML, MediaType.TEXT_PLAIN, MediaType.APPLICATION_JSON})
	public MxModel list(@PathParam("server") String server,
	                    @PathParam("domain") String domain,
	                    @PathParam("mbean") String mbean) {
		try {
			MBeanServerConnection serverConnection = serverConnectionProvider.getConnection(server);

			List<MBeanAttributeInfo>
					result =
					new ArrayList<MBeanAttributeInfo>(Arrays.asList(serverConnection.getMBeanInfo(new ObjectName(
							domain + ":" + mbean)).getAttributes()));

			Iterator<MBeanAttributeInfo> i = result.iterator();

			// Filters unreadable attributes
			while (i.hasNext()) {
				if (!i.next().isReadable()) {
					i.remove();
				}
			}

			Collections.sort(result, new Comparator<MBeanAttributeInfo>() {
				public int compare(MBeanAttributeInfo o1, MBeanAttributeInfo o2) {
					return o1.getName().compareTo(o2.getName());
				}
			});
			return new MxListModel("attributes", "attribute", result);
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (IntrospectionException e) {
			throw new RuntimeException(e);
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		}
	}

	@Path("{attribute}")
	@GET
	@Produces({MediaType.TEXT_HTML, MediaType.TEXT_PLAIN, MediaType.APPLICATION_JSON})
	public MxModel get(@PathParam("server") String server,
	                   @PathParam("domain") String domain,
	                   @PathParam("mbean") String mbean,
	                   @PathParam("attribute") String attribute,
	                   @QueryParam("ok") @DefaultValue("false") boolean isPostback) {
		Map<String, Object> model = new HashMap<String, Object>();

		try {
			MBeanServerConnection serverConnection = serverConnectionProvider.getConnection(server);

			MBeanAttributeInfo info = null;
			for (MBeanAttributeInfo i : serverConnection.getMBeanInfo(new ObjectName(domain + ":" + mbean))
					.getAttributes()) {
				if (i.getName().equals(attribute)) {
					info = i;
				}
			}

			Object value = serverConnection.getAttribute(new ObjectName(domain + ":" + mbean), attribute);

			model.put("server", server);
			model.put("domain", domain);
			model.put("mbean", mbean);
			model.put("attribute", info);
			if (isPostback) {
				model.put("ok", true);
			}
			if (value == null) {
				model.put("value", "<null>");
			} else if (value.getClass().isArray()) {
				templateName = "array-attribute";
				model.put("items", value);
			} else if (value instanceof CompositeData) {
				templateName = "composite-attribute";
				model.put("attribute", value);
			} else if (value instanceof TabularData) {
				templateName = "tabular-attribute";
				model.put("attribute", value);
			} else {
				model.put("value", value);
			}

		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (AttributeNotFoundException e) {
			throw new RuntimeException(e);
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (MBeanException e) {
			Exception targetException = e.getTargetException();
			if (targetException instanceof RuntimeErrorException) {
				throw new RuntimeException(targetException.getCause());
			}
			log.warn("Error accessing attribute", e);
			model.put("value", e.getTargetException().getCause().getMessage());
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		} catch (IntrospectionException e) {
			throw new RuntimeException(e);
		} catch (RuntimeException e) {
			model.put("value", e.getMessage());
			log.warn("Error accessing attribute", e);
		}
		return new MxItemModel(attribute, true, templateName, model);
	}

	@Path("{attribute}")
	@POST
	public Response post(@PathParam("server") String server,
	                     @PathParam("domain") String domain,
	                     @PathParam("mbean") String mbean,
	                     @PathParam("attribute") String attributeName,
	                     @FormParam("value") String value) {
		MBeanServerConnection serverConnection = serverConnectionProvider.getConnection(server);

		try {

			String type = String.class.getName();
			for (MBeanAttributeInfo info : serverConnection.getMBeanInfo(new ObjectName(domain + ":" + mbean))
					.getAttributes()) {
				if (info.getName().equals(attributeName)) {
					type = info.getType();
				}
			}

			Object attribute = new ValueParser().parse(value, type);

			if (attribute != null) {
				serverConnection
						.setAttribute(new ObjectName(domain + ":" + mbean), new Attribute(attributeName, attribute));
			}

			return Response.seeOther(uriInfo.getRequestUriBuilder().queryParam("ok", true).build()).build();
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (AttributeNotFoundException e) {
			throw new RuntimeException(e);
		} catch (InvalidAttributeValueException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (MBeanException e) {
			throw new RuntimeException(e);
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (IntrospectionException e) {
			throw new RuntimeException(e);
		}

	}
}
