/*
 * Copyright 2012 Jared Bunting, Peach Jean Solutions <jared.bunting@peachjean.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package net.peachjean.restmx.console.resource;

import net.peachjean.restmx.console.model.MxItemModel;
import net.peachjean.restmx.console.model.MxListModel;
import net.peachjean.restmx.console.model.MxModel;
import net.peachjean.restmx.server.ServerConnectionProvider;
import net.sf.json.JSONSerializer;

import javax.management.*;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.io.IOException;
import java.util.*;

@Path("/servers/{server}/domains/{domain}/mbeans/{mbean}/operations")
public class OperationResource {

	@Context
	private ServerConnectionProvider serverConnectionProvider;

	@Context
	private UriInfo uriInfo;

	@GET
	@Produces({MediaType.TEXT_HTML, MediaType.TEXT_PLAIN, MediaType.APPLICATION_JSON})
	public MxModel getList(@PathParam("server") String server,
	                       @PathParam("domain") String domain,
	                       @PathParam("mbean") String mbean) {

		try {
			MBeanServerConnection serverConnection = serverConnectionProvider.getConnection(server);

			List<Map<String, Object>> result = new ArrayList<Map<String, Object>>();

			for (MBeanOperationInfo info : serverConnection.getMBeanInfo(new ObjectName(domain + ":" + mbean))
					.getOperations()) {
				Map<String, Object> pair = new HashMap<String, Object>();

				// Skip getters
				Object role = info.getDescriptor().getFieldValue("role");
				if (role == null && info.getName().startsWith("get") && info.getSignature().length == 0 || "getter"
						.equals(role)) {
					continue;
				}

				StringBuilder sb = new StringBuilder(info.getName() + "(");
				boolean first = true;
				for (MBeanParameterInfo p : info.getSignature()) {
					if (!first) {
						sb.append(",");
					} else {
						first = false;
					}
					sb.append(p.getType());
				}
				sb.append(")");

				pair.put("operation", info);
				pair.put("declaration", sb.toString());
				result.add(pair);
			}

			Collections.sort(result, new Comparator<Map<String, Object>>() {
				public int compare(Map<String, Object> o1, Map<String, Object> o2) {
					return ((String) o1.get("declaration")).compareTo((String) o2.get("declaration"));
				}
			});

			return new MxListModel("operation", "operations", result);

		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (IntrospectionException e) {
			throw new RuntimeException(e);
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		}

	}

	@Path("{declaration}")
	@GET
	@Produces({MediaType.TEXT_HTML, MediaType.TEXT_PLAIN, MediaType.APPLICATION_JSON})
	public MxModel get(@PathParam("server") String server,
	                   @PathParam("domain") String domain,
	                   @PathParam("mbean") String mbean,
	                   @PathParam("declaration") String declaration,
	                   @QueryParam("ok") @DefaultValue("false") boolean isPostback) {
		String operation = declaration.split("\\(")[0];

		String signature = declaration.split("\\(").length > 1 ?
				(declaration.split("\\(")[1].split("\\)").length > 0 ?
						declaration.split("\\(")[1].split("\\)")[0] : "") : "";

		Map<String, Object> model = new HashMap<String, Object>();

		model.put("server", server);
		model.put("domain", domain);
		model.put("mbean", mbean);
		model.put("operation", getOperation(serverConnectionProvider
				.getConnection(server), domain, mbean, operation, signature));
		if (isPostback) {
			model.put("ok", true);
		}

		return new MxItemModel(declaration, true, "operation", model);
	}

	@Path("{declaration}")
	@POST
	@Produces({MediaType.TEXT_PLAIN})
	public Response postForText(@PathParam("server") String server,
	                            @PathParam("domain") String domain,
	                            @PathParam("mbean") String mbean,
	                            @PathParam("declaration") String declaration,
	                            @FormParam("param") String[] stringParams) {
		return this.doPost(server, domain, mbean, declaration, stringParams, new ResultHandler() {
			@Override
			public Response getResponse(final Object result) {
				return Response.ok(result.toString()).build();
			}
		});
	}

	@Path("{declaration}")
	@POST
	@Produces({MediaType.APPLICATION_JSON})
	public Response postForJson(@PathParam("server") String server,
	                            @PathParam("domain") String domain,
	                            @PathParam("mbean") String mbean,
	                            @PathParam("declaration") String declaration,
	                            @FormParam("param") String[] stringParams) {
		return this.doPost(server, domain, mbean, declaration, stringParams, new ResultHandler() {
			@Override
			public Response getResponse(final Object result) {
				return Response.ok(JSONSerializer.toJSON(result).toString()).build();
			}
		});
	}


	private Response doPost(String server,
	                        String domain,
	                        String mbean,
	                        String declaration,
	                        String[] stringParams,
	                        ResultHandler resultHandler) {
		String operation = declaration.split("\\(")[0];

		String[] signature = declaration.split("\\(").length > 1 ? (
				declaration.split("\\(")[1].split("\\)").length > 0 ?
						declaration.split("\\(")[1].split("\\)")[0].split(",") :
						new String[]{}) : new String[]{};

		MBeanServerConnection serverConnection = serverConnectionProvider.getConnection(server);

		try {

			Object[] params = new Object[signature.length];

			ValueParser parser = new ValueParser();
			for (int i = 0; i < stringParams.length; i++) {
				params[i] = parser.parse(stringParams[i], signature[i]);
			}

			Object result = serverConnection.invoke(new ObjectName(domain + ":" + mbean), operation, params, signature);

			if (result != null) {
				return resultHandler.getResponse(result);
			} else {
				return Response.seeOther(uriInfo.getRequestUriBuilder().queryParam("ok", true).build()).build();
			}
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (MBeanException e) {
			throw new RuntimeException(e);
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private MBeanOperationInfo getOperation(MBeanServerConnection server, String domain, String mbean,
	                                        String operationName, String signature) {
		try {

			for (MBeanOperationInfo i : server.getMBeanInfo(new ObjectName(domain + ":" + mbean)).getOperations()) {

				StringBuilder sb = new StringBuilder();
				boolean first = true;
				for (MBeanParameterInfo p : i.getSignature()) {
					if (!first) {
						sb.append(",");
					} else {
						first = false;
					}
					sb.append(p.getType());
				}

				if (i.getName().equals(operationName)) {
					if (sb.toString().equals(signature)) {
						return i;
					}
				}
			}
			return null;
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (InstanceNotFoundException e) {
			throw new RuntimeException(e);
		} catch (MalformedObjectNameException e) {
			throw new RuntimeException(e);
		} catch (ReflectionException e) {
			throw new RuntimeException(e);
		} catch (IntrospectionException e) {
			throw new RuntimeException(e);
		}
	}

	private interface ResultHandler {
		Response getResponse(Object result);
	}
}
